#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Dec  6 08:55:35 2019

@author: wolterse
"""

import numpy as np
import seaborn as sns
import matplotlib.pyplot as plt


from mpl_toolkits.basemap import Basemap


from osgeo import gdal

import matplotlib.colors as colors
import cartopy.crs as ccrs
import rasterio
import glob

def im_plot(data,lats,lons,outPNG,figsize=[12,15],drawcoastlines=True,
            title='',vmin=None,vmax=None,cmap='jet',bounds=None,plotlegend=True,extend='max', 
            leg_label=''):
    """ 
    Plot 2D map of georeferenced image data.
    Required input: 
        data: data array to be plotted
        lats: array of latitudes (or [minlat,maxlat])
        lons: array of longitudes (or [minlon,maxlon])
        outPNG: outfilename
    Optional input:
        figsize: size of the plot (default=[12,15])
        drawcoastlines (default=True)
        title: plot title
        vmin: minimum data value to be plotted (default=None)
        vmax: maximum data value to be plotted (default=None)
        cmap: colormap (default 'jet'), can be ListedColormap
        bounds: boundaries for cmap array (default=None)
        plotlegend: True/False (default=True)
        extend: extend colorbar 'neither','both','min' or 'max (default='max')
        leg_label: legend label for the colorbar
    """
    
    fig=plt.figure(figsize=figsize)
    ax=plt.axes(projection=ccrs.PlateCarree())
    #img_extent=np.array(([lons[0],lons[-1],lats[-1],lats[0]]))
    img_extent = np.array(region)
    ax.set_extent(img_extent,crs=ccrs.PlateCarree())
    
    if drawcoastlines==True:
        ax.stock_img()
        ax.coastlines(resolution='50m')
    if np.size(bounds) > 1:
        if isinstance(cmap,str):
            norm=colors.BoundaryNorm(bounds,ncolors=256)
        else:
            norm=colors.BoundaryNorm(bounds,ncolors=cmap.N)
    else:
        norm=None
        bounds=None
        
#    img = ax.imshow(data,origin='upper',extent=img_extent,transform=ccrs.PlateCarree(),cmap=cmap, vmin=vmin,vmax=vmax)
    img = ax.imshow(data, cmap=cmap, extent=img_extent, vmin=vmin,vmax=vmax)
    plt.title(title, fontsize=20)
    
    if plotlegend==True:
        #sm = plt.cm.ScalarMappable(cmap=cmap, norm=norm)
        #sm.set_array([])
        if np.size(bounds) > 1:
            plt.colorbar(img,ax=ax,cmap=cmap,boundaries=bounds,ticks=bounds[1:-1],
                         fraction=0.023,pad=0.02)
        else:
            cb = plt.colorbar(img, ax=ax,cmap=cmap,
                         fraction=0.023,pad=0.02)
            cb.ax.set_ylabel(leg_label)
            cb.ax.tick_params(labelsize=14)

    fig.savefig(outPNG,bbox_inches='tight')
    
###############################################################################

#dir_img   = '/data/TERRASCOPE/s5p_dev/output/L3_CH4_TM_V1/2020/S5P_OFFL_L3_CH4_TM_20200801_V100'
dir_img   = '/data/TERRASCOPE/s5p_dev/output/L3_CH4_TM_V1/2022/S5P_OFFL_L3_CH4_TM_20220101_V100'
dir_plt   = '/data/users/Private/lwu/s5p_2022_1215'
t_aggr    = 'TM'
region    = [-180, 180, -90, 90]
split_dir = dir_img.split('/')[5]
atm_comp  = split_dir.split('_')[1]
period    = split_dir.split('_')[2]
str_title = 'S-5P/TROPOMI column integrated ' + atm_comp + ' ' + period

if atm_comp == 'CO':
    vmin  = 0.
    vmax  = 150.
    label = '[ppbv]'
    scale = 1.
elif atm_comp == 'CH4':
    vmin  = 1650
    vmax  = 1950.
    label = '[ppbv]'
    scale = 1.
else:
    vmin  = 0.
    vmax  = 100.
    label = '[$\mu$mol m$^{-2}$]'
    scale = 1000000.
#get the monthly/daily/week/weekend/Sat/Sun file to plot
lst_files = glob.glob(dir_img + '/*' + t_aggr + '*.tif')
print('lst_files ',lst_files )
src   = rasterio.open(lst_files[0])
lon   = src.bounds[0] + np.arange(0, src.width) * 0.05
lat   = src.bounds[3] - np.arange(0, src.height) * 0.05

print('d_s5p',src.read(1))
#d_s5p = src.read(1) * src.scales + src.offsets
#print('d_s5p',d_s5p)
#
#print('scale',src.scales)
#print('offset',src.offsets)
#print('offset',src.scales)
#print('ds',ds)
#print('ds',ds.offsets)
#data = ds.ReadAsArray()
#print('data',data)

ds    = gdal.Open(lst_files[0])
data  = ds.ReadAsArray()
data2 = np.ma.masked_values(data, -9999) 
#data2 = np.ma.masked_values(src.read(1), -9999) 

d_s5p =  data2* src.scales + src.offsets

print('d_s5p ',d_s5p)
print('vmin', vmin)
print('vmax',vmax)
###fig, axs = plt.subplots(1, 1, tight_layout=True)
###axs.hist(d_s5p, range=(vmin, vmax))
###fig.savefig('hist_ch4.png',bbox_inches='tight')


i_lon = np.where((lon > region[0]) & (lon <= region[1]))
i_lat = np.where((lat > region[2]) & (lat <= region[3]))


map = Basemap(projection='cyl')
 
map.drawmapboundary()
map.drawcoastlines()


#x = np.linspace(0, map.urcrnrx, data.shape[1])
#y = np.linspace(0, map.urcrnry, data.shape[0])

xx, yy = np.meshgrid(lon, lat)

cp = map.contourf(xx, yy, d_s5p, levels=np.linspace(vmin, vmax, 300),  cmap='Spectral_r', extend='min')
plt.colorbar(cp, location='bottom')

plt.savefig('test.png')
#plt.show()

#plot the image 
#im_plot(d_s5p[i_lat[0][0]:i_lat[0][-1],i_lon[0][0]:i_lon[0][-1]],lat,lon,dir_plt + '/' + dir_img.split('/')[-1] + '.png',figsize=[12,15],drawcoastlines=True,
#            title=str_title,vmin=vmin,vmax=vmax,cmap='Spectral_r',bounds=None,
#            plotlegend=True,extend=region, leg_label=label)