#!/bin/bash
input_dir=$1
output_dir_nc=$2
output_dir_tif=$3
composit_type=$4
product_type=$5
processing_mode=$6
date_lst_file=$7
summary_file=$8
job_name=${9}
version=${10}
output_mount_dir=${11}
also_use_RPRO=${12}


#some example commands
# ./run_workflow_runner.sh /data/MTDA/CGS_S5P/CGS_S5P_L2_CH4 /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TD_V1 /data/TERRASCOPE/s5p_dev/output/L3_CH4_TD_V1 TD CH4 OFFL 20200101 20200131 /data/TERRASCOPE/s5p_dev/summary_L3_CH4_TD.json S5P_L3_CH4_TD_test 100 /data/TERRASCOPE/s5p_dev

export SPARK_HOME=/opt/spark3_2_0/
export PYSPARK_PYTHON=python3.8

DOCKER_IMAGE='vito-docker-private.artifactory.vgt.vito.be/s5p_harp1.16'

DIRNAME=`dirname $0`
DIRNAME="${DIRNAME}/.."
cd $DIRNAME
ls s5pbira

keyfile="/etc/security/keytabs/mep_cgs.keytab"
kinit -kt $keyfile mep_cgs


#  --conf spark.yarn.appMasterEnv.PYSPARK_PYTHON=/usr/bin/python3.6 \
#  --conf spark.yarn.appMasterEnv.SPARK_HOME=/usr/hdp/current/spark2-client \


zip -r "/tmp/s5pbira2.zip" s5pbira/*.py s5pbira/config/*keytab  && \
  spark-submit --master yarn --queue mep_cgs --deploy-mode cluster \
  --name "${job_name}" \
  --conf spark.yarn.appMasterEnv.YARN_CONTAINER_RUNTIME_TYPE=docker \
  --conf spark.executorEnv.YARN_CONTAINER_RUNTIME_TYPE=docker \
  --conf spark.yarn.appMasterEnv.YARN_CONTAINER_RUNTIME_DOCKER_IMAGE=${DOCKER_IMAGE} \
  --conf spark.executorEnv.YARN_CONTAINER_RUNTIME_DOCKER_IMAGE=${DOCKER_IMAGE} \
  --conf spark.yarn.appMasterEnv.YARN_CONTAINER_RUNTIME_DOCKER_MOUNTS=/var/lib/sss/pipes:/var/lib/sss/pipes:rw,/usr/hdp/current/:/usr/hdp/current/:ro,/etc/hadoop/conf/:/etc/hadoop/conf/:ro,/etc/krb5.conf:/etc/krb5.conf:ro,/data/MTDA/CGS_S5P:/data/MTDA/CGS_S5P:ro,${output_mount_dir}:${output_mount_dir}:rw \
  --conf spark.executorEnv.YARN_CONTAINER_RUNTIME_DOCKER_MOUNTS=/var/lib/sss/pipes:/var/lib/sss/pipes:rw,/usr/hdp/current/:/usr/hdp/current/:ro,/etc/hadoop/conf/:/etc/hadoop/conf/:ro,/etc/krb5.conf:/etc/krb5.conf:ro,/data/MTDA/CGS_S5P:/data/MTDA/CGS_S5P:ro,${output_mount_dir}:${output_mount_dir}:rw \
  --conf spark.executorEnv.INCLUDE=/usr/include/ --conf spark.executorEnv.LD_LIBRARY_PATH=$LD_LIBRARY_PATH:/opt/harp/lib:/opt/coda/lib/ \
  --conf spark.yarn.appMasterEnv.PYSPARK_DRIVER_PYTHON=/usr/bin/python3.8 \
  --conf spark.executorEnv.PYSPARK_PYTHON=/usr/bin/python3.8 \
  --conf spark.shuffle.service.enabled=true \
  --conf spark.dynamicAllocation.enabled=true \
  --conf spark.dynamicAllocation.minExecutors=1 \
  --conf spark.yarn.submit.waitAppCompletion=true \
  --conf spark.yarn.maxAppAttempts=1 \
  --conf spark.executor.memoryOverhead=4G \
  --conf spark.ui.view.acls.groups=vito \
  --conf spark.modify.acls.groups=vito \
  --principal "mep_cgs@VGT.VITO.BE" \
  --keytab $keyfile \
  --executor-memory 512M --py-files /tmp/s5pbira2.zip s5pbira/workflow_runner.py \
  --input_dir ${input_dir} \
  --output_dir_tif ${output_dir_tif} \
  --output_dir_nc ${output_dir_nc} \
  --product_type ${product_type} \
  --composit_type ${composit_type} \
  --processing_mode ${processing_mode} \
  --date_lst_file ${date_lst_file} \
  --version ${version} \
  --summary_file ${summary_file} \
  --cluster \
  --also_use_RPRO ${also_use_RPRO}
