#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import logging
logformat='%(asctime)s %(name)s %(levelname)-8s %(message)s'
logtfmt='%y-%m-%d %H:%M:%S'
#set the basis configuration for logging to the console 
logging.basicConfig(level=logging.INFO,format=logformat,datefmt=logtfmt)
logger=logging.getLogger('BIRA')
testlogger=logging.getLogger('TEST')
testlogger.setLevel(logging.DEBUG)
operlogger=logging.getLogger('TERRA')
operlogger.setLevel(logging.ERROR)
import argparse,os,fnmatch,glob,multiprocessing,subprocess,re
import numpy as np
import datetime as dt

__version__='1.0'

_MAXCPU=multiprocessing.cpu_count()
_MAXBINCOUNT=int('fffff',16) #max number of files I can track in the binning filename, the 5 characters from the orbit number   
_LENS5PID=len('S5P_RPRO_L2__CO____')
_NOFILEFILTER=lambda _f: True
_AVERAGE_THRESHOLD=dict(TM=1,TY=1,m5=1,m6=1,ms=1,mw=1)#dict(TD=7,TM=15,TY=8,m5=3,m6=3,ms=2*3,mw=5*3) #don't created these averages if the number of input files is below the threshold: at least 7 orbits in a daily file,...

#iddict will relate different processor versions... if they match than they can be binned to one file
#eg a RPRO 01_010202 can be combined with an OFFL 01_010202 file, or a TB OFFL 01_010202 with a TB 01_010300 to one TD file labelled O(FFL) 01_010202 ...  
iddict={
  ('S5P_OFFL_L3{avg_level}CO____','01_010202'): [('S5P_OFFL_L2__CO____','01_010202'),('S5P_OFFL_L2__CO____','01_010[3-9][0-9][0-9]'),('S5P_RPRO_L2__CO____','01_010202'),('S5P_RPRO_L2__CO____','01_010[3-9][0-9][0-9]')],
  ('S5P_OFFL_L3{avg_level}CO____','02_020200'): [('S5P_OFFL_L2__CO____','02_020[2-9][0-9][0-9]'),('S5P_RPRO_L2__CO____','02_020[2-9][0-9][0-9]')],
  ('S5P_NRTI_L3{avg_level}CO____','000000'): [('S5P_NRTI_L2__CO____','[0-9]*')],
  ('S5P_OFFL_L3{avg_level}NO2___','01_010202'): [('S5P_OFFL_L2__NO2___','01_010202'),('S5P_OFFL_L2__NO2___','01_0103[0-9][0-9]'),('S5P_RPRO_L2__NO2___','01_010202'),('S5P_RPRO_L2__NO2___','01_0103[0-9][0-9]')],
  ('S5P_OFFL_L3{avg_level}NO2___','01_010400'): [('S5P_OFFL_L2__NO2___','01_0104[0-9][0-9]')],
  ('S5P_OFFL_L3{avg_level}NO2___','02_020200'): [('S5P_OFFL_L2__NO2___','02_020[2-9][0-9][0-9]'),('S5P_RPRO_L2__NO2___','02_020[2-9][0-9][0-9]')],  
  ('S5P_NRTI_L3{avg_level}NO2___','000000'): [('S5P_NRTI_L2__NO2___','[0-9]*')],
  ('S5P_OFFL_L3{avg_level}CH4___','01_010202'): [('S5P_OFFL_L2__CH4___','01_010202'),('S5P_OFFL_L2__CH4___','01_010[3-9][0-9][0-9]'),('S5P_RPRO_L2__CH4___','01_010202'),('S5P_RPRO_L2__CH4___','01_010[3-9][0-9][0-9]')],
  ('S5P_OFFL_L3{avg_level}CH4___','02_020200'): [('S5P_OFFL_L2__CH4___','02_020[2-9][0-9][0-9]'),('S5P_RPRO_L2__CH4___','02_020[2-9][0-9][0-9]')],    
  ('S5P_OFFL_L3{avg_level}HCHO__','01_010105'): [('S5P_OFFL_L2__HCHO__','01_01010[56789]'),('S5P_OFFL_L2__HCHO__','01_02010[0-9]'),('S5P_RPRO_L2__HCHO__','01_010105')],
  ('S5P_OFFL_L3{avg_level}HCHO__','02_020200'): [('S5P_OFFL_L2__HCHO__','02_020[2-9][0-9][0-9]'),('S5P_RPRO_L2__HCHO__','02_020[2-9][0-9][0-9]')]
}
#for public data we need a different iddict
iddict_vito={
  ('S5P_OFFL_L3{avg_level}CO____','01_010202'): [('S5P_OFFL_L2__CO____','01_010202'),('S5P_OFFL_L2__CO____','01_010100'),('S5P_OFFL_L2__CO____','01_010200'),('S5P_OFFL_L2__CO____','01_010[3-9][0-9][0-9]'), ('S5P_OFFL_L2__CO____','0[23]_020[2-9][0-9][0-9]'), ('S5P_RPRO_L2__CO____','01_010202')],
  ('S5P_NRTI_L3{avg_level}CO____','000000'): [('S5P_NRTI_L2__CO____','[0-9]*')],
  ('S5P_OFFL_L3{avg_level}NO2___','01_010202'): [('S5P_OFFL_L2__NO2___','01_010202'),('S5P_OFFL_L2__NO2___','01_010200'),('S5P_OFFL_L2__NO2___','01_010[34][0-9][0-9]'),('S5P_OFFL_L2__NO2___','0[23]_020[2-9][0-9][0-9]'), ('S5P_RPRO_L2__NO2___','01_010202'),('S5P_RPRO_L2__NO2___','01_0103[0-9][0-9]')],
  ('S5P_NRTI_L3{avg_level}NO2___','000000'): [('S5P_NRTI_L2__NO2___','[0-9]*')]
}

iddict_match={vi:k for k,v in iddict.items() for vi in v}
iddict_vito_match={vi:k for k,v in iddict_vito.items() for vi in v}


def idmap(f):
  """get a unique id from a L2/L3 file, consists of S5P_OFFL_L3TMCO____ and collection_processor (OFFL,RPRO) or orbitnumber(NRT) """
  p='_'.join(f.split('_')[-3:-1]) if not 'S5P_NRTI_L2__' in f[:_LENS5PID] else f.split('_')[-4]  
  #this is a hardcoded fix: 010301 was only there to fill up gaps and 010302 only appears on 2018040[4,5], such a reprocessing is not foreseen ... should this exception be extended to other KNMI processor products? not for NO2, yes for CH4... TODO 
  if p in ('01_010301','01_010302') and f[:_LENS5PID] in ('S5P_RPRO_L2__CO____','S5P_RPRO_L2__CH4___'): p='01_010202'
  return (f[:_LENS5PID],p) #helper function to get type,processor from filename

#functions to help create bins
getmonthbin=lambda t: dt.date(t.year,t.month,1)
getyearbin=lambda t: dt.date(t.year,1,1)

#get datetime from file name component
gettime=lambda s: dt.datetime.strptime(s,'%Y%m%dT%H%M%S') 

def check_existing_time_window(f,stime,etime): 
  """function to check existing L3 file time range against given time range"""
  try: fstime,fetime=map(gettime,f.split('_')[-6:-4])
  except Exception as tfe: logger.error('Could not determine time from file %s: %s'%(f,repr(tfe)));raise tfe
  return (stime<=fstime and etime>=fetime)

def run_job(j):
  return subprocess.run(j,shell=True,stdout=subprocess.PIPE,stderr=subprocess.PIPE)#  => for older versions of python3 at Bira, otherwise use j,shell=True,capture_output=True)#

# o = run_job('harpmerge -v')
# if o.returncode: raise ValueError('Error in calling harpmerge to determine harp version: %s'%o)
# else:
#   harpversion=tuple(map(int,re.sub('[^0-9.]','',o.stdout.decode('ascii').split('\n')[0]).split('.')));
#   logger.debug('Working with HARP version %s'%'.'.join(map(str,harpversion)))
harpversion=(1,10,0)
if harpversion<(1,9,2): raise ValueError('HARP version should be above 1.9.2 for this version of the script')

def run_folder(data_folder,output_folder,latstep=0.05,lonstep=0.05,filefilter=_NOFILEFILTER,avg_level_filter='*',simulation=True,npool=0,automode=False,VITO=False,box=None,logger=logger,extra_post_actions=None,**kargs): 
  """function that calls the HARP binning tool with the appropriate arguments
  
  data_folder=data folder to process, will loop over all subfolders and process all files per subfolder (we have organised L2 data in daily folders)
  
  output_folder=output folder, willl contain the binned files
  
  simulation=True, set False to execute the binning operations
  automode=False, will not recreate existing files if True
  npool=0, use >0 for multiprocessing
  filefilter=function applied file lists in a a folder
  VITO=False, use True for VITO environment, also changes the iddict mapping
  logger=logger instance
  box=tuple for lat lon box
  extra_post_actions=add extra ap arguments (only when -ar is used for binning)
  avg_level_filter='*', filter on which L3 levels to create
  kargs=passed on to harpmerge
  
  
  t.run_folder('/bira-iasb/data/SATELLITE/S5P/TROPOMI/L2/ESA_HUB/OFL/CH4/01_01/2019',output_folder='/home/bavol/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__CH4/%Y/%m',latstep=0.2,lonstep=0.2,simulation=False,automode=True) #creates daily files
  
  t.run_folder('/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__CH4',output_folder='/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__CH4/%Y',latstep=0.2,lonstep=0.2,simulation=False,automode=True)
  
  t.run_folder('/bira-iasb/data/SATELLITE/S5P/TROPOMI/L2/ESA_HUB/OFL/AER_LH/01_01/2018/12',output_folder='/home/bavol/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__AER_LH/%Y/%m',latstep=0.2,lonstep=0.2,simulation=False,automode=False)
  t.run_folder('/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__AER_LH',output_folder='/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__AER_LH/%Y',latstep=0.2,lonstep=0.2,simulation=False,automode=True)
  t.run_folder('/bira-iasb/data/SATELLITE/S5P/TROPOMI/L2/ESA_HUB/OFL/HCHO/01_01/2019/01',output_folder='/home/bavol/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__HCHO/%Y/%m',latstep=0.2,lonstep=0.2,simulation=False,automode=False)
  t.run_folder('/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__AER_LH',output_folder='/bira-iasb/projects/FTIR/validation/s5p/binned_files/S5P_OFFL_L2__AER_LH/%Y',latstep=0.2,lonstep=0.2,simulation=False,automode=True)
  """
  
  #### Definitions
  def idmatch(s5pid,p,M=iddict_vito_match if VITO else iddict_match):
    """function that gets an s5pid,processorversion and returns the value from iddict that matches to it...

    idmatch('S5P_NRTI_L2__CO____','010302')->('S5P_NRTI_L3{avg_level}CO____', '000000')
    idmatch('S5P_OFFL_L3TMCO____','010302')->('S5P_OFFL_L3{avg_level}CO____', '010302')"""
    if not len(s5pid)==_LENS5PID: raise ValueError('Invalid identifier %s should have length %d'%(s5pid,_LENS5PID))
    k = list(filter(lambda i_p : i_p[0]==s5pid and fnmatch.fnmatch(p,i_p[1]),M.keys())) or [None]
    if len(k)>1: raise ValueError('Retrieved more than one match for %s, %s from %s: %s'%(s5pid,p,M,k))
    if '_L2__' in s5pid and not k[0]: logger.warning('Processing unknown processor version %s,%s'%(s5pid,p))
    return M.get(k[0],(s5pid[:9]+'L3{avg_level}'+s5pid[13:],p))
  keep='keep(%s,latitude_bounds,longitude_bounds,datetime_start)' #variables to keep when loading S5P file: only these are loaded into memory, %s is filled later on with the right target variable
  exclude_default='exclude(latitude_bounds_weight,longitude_bounds_weight,latitude,longitude)'
  #S5P_RPRO_L2__CO_____20180531T022053_20180531T040421_03259_01_010202_20190206T143349.nc
  #construct arrays with bin boundaries and add to oper
  #setup up eg bin_spatial(3601,-90,0.05,7201,-180,0.05)
  if box: 
    logger.warning('Experimental feature')
    oper_box=('latitude>=%5.2f;latitude<=%5.2f;longitude_range(%5.2f,%5.2f)'%(box[0]-5*latstep,box[1]+5*latstep,box[2]-5*lonstep,box[3]+5*lonstep)).split(';') #(51.5,62.5,-94.5,-56.5)(-90,90,-180,180)
    oper_bin='bin_spatial(%d,%.2f,%.2f,%d,%.2f,%.2f)'%(int((box[1]-box[0])/latstep)+1,box[0],latstep,int((box[3]-box[2])/lonstep+1),box[2],lonstep) 
  else: oper_box=[];oper_bin='bin_spatial(%d,-90,%.2f,%d,-180,%.2f)'%(int(180/latstep)+1,latstep,int(360/lonstep+1),lonstep) 

  extra_harp_args=''.join(['--%s %s '%(k.replace('hdf5_compression','hdf5-compression'),v) for k,v in kargs.items()])
  logger.debug('Extra harpmerge arguments: %s'%extra_harp_args)

  if extra_post_actions and harpversion<(1,10): raise ValueError('extra_post_actions argument is only allowed for harp>=1.10')
  logger.debug('Extra HARP post bin actions -ap: %s'%extra_post_actions)

  simulation=bool(simulation)
  jobs=[]
  
  def _checkoutput(r,fullout):
    """helper function to check the binned file and deal with error codes"""
    if not hasattr(r,'returncode'): err=1 #some kind of exception occured, captured before
    else: err=r.returncode
    if err==2 and b'WARNING: merged product is empty\n'==r.stderr: pass
    elif err==-9: err=1;logger.error('Not enough memory for %s'%fullout)
    elif err==0 and not os.path.isfile(fullout): err=1
    elif err==0: #just check to be absolutely sure! 
      checkres=run_job('harpcheck '+fullout)
      err=checkres.returncode
      if err: os.remove(fullout);logger.info('Removed %s: %s'%(fullout,checkres.stdout+checkres.stderr))      
    else: err=1;logger.warning('Unknown returncode for harpmerge for %s: %s'%(fullout,r))
    
    #deal with the different error codes
    if err==1: logger.error('HARP failure for %s'%fullout)
    elif err==2: open(fullout[:-2]+'0c','w');logger.info('\tcreated empty merge file for %s'%fullout) #0c means no pixels passed the filter, the creation of this file avoids redoing the same job each time the script is executed
    else: logger.info('\tcreated %s'%fullout)

  
  #filter function to get sublists: input is a file+ extra arguments
  dayfilter=lambda f,d: gettime(f.split('_')[-6]).date()==d
  monthfilter=lambda f,d: getmonthbin(gettime(f.split('_')[-6]).date())==d
  yearfilter=lambda f,d: getyearbin(gettime(f.split('_')[-6]).date())==d
  weekdayfilter=lambda f,m,wkdays: getmonthbin(gettime(f.split('_')[-6]).date())==m and gettime(f.split('_')[-6]).weekday() in wkdays
  filemap=dict()
  
  #### Run
  #if not simulation and not os.path.isdir(output_folder): os.makedirs(output_folder)
  for _p,_ds,_fs in os.walk(data_folder,topdown=False,followlinks=True): 
    if not VITO: _fs=list(filter(lambda f: filefilter(f) and f[0]!='.' and f[-3:]=='.nc',_fs));[filemap.setdefault(f,_p) for f in _fs]
    else: #VITO setup: files are in subfolders
      _fs=list(filter(lambda f: filefilter(f) and f[0]!='.' and glob.glob(os.path.join(_p,f,'*.nc')),_ds)) #use directories list in os.walk that contains nc files
      [filemap.setdefault(f+'.nc',os.path.join(_p,f)) for f in _fs]; #remember location of files
      _fs=list(map(lambda x: x+'.nc',_fs)) #add extension
    if not _fs: continue
    #determine the processor versions in the list of files
    processors=map(idmap,_fs)
    processors_set=sorted(set(processors))
    if len(processors_set)>1: logger.info('Found multiple processors or binning groups in %s'%_p)
    #clean up input folder ... ESA did some strange things in the RPRO time series
    l2_fs=list(filter(lambda f: '_L2__' in idmap(f)[0][:_LENS5PID],_fs)) # in ('S5P_OFFL_L2__','S5P_RPRO_L2__')
    if len(l2_fs): 
      #detect if L2 orbits are in _fs multiple times ...keep highest processor version/generation date occurs in eg RPRO/CO/01_01/2018/06/17 or 2018/04/30
      orbits_list=list(map(lambda f: '_'.join(f.split('_')[:-3]),l2_fs));orbits_set=set(orbits_list) #orbits_list uses the entire filename up to the orbit number to identify a file (NRT files come in granules and need stime,etime) 
      toremove=[]
      for o in orbits_set:
        if orbits_list.count(o)>1:
          matching_ids,matching_filenames=list(zip(*sorted([(idmatch(*idmap(f)),f) for f,_o in zip(l2_fs,orbits_list) if _o==o],key=lambda IF: IF[1]))) # zipping list of ((news5pid,newprocessor),filename)  sorted on filename... 
          for matchingid in set(matching_ids):
            if matching_ids.count(matchingid)>1: toremove.extend([f for _id,f in zip(matching_ids,matching_filenames) if _id == matchingid][:-1])
      [(_fs.remove(_f),logger.debug('Removed file %s'%os.path.basename(_f))) for _f in toremove]   
    if not _fs: continue
    #recalculate processor set
    processors=map(idmap,_fs)
    processors_set=sorted(set(processors))
    #list of files is now ready to determine the to be merged sublists
    for t,p in processors_set:
      oper=[] #list of operations arguments to parse to HARP 
      option='' #harp options, only for ch4 
      fs_p=sorted(filter(lambda f: idmap(f)==(t,p), _fs),key=lambda ff: ff.split('_')[-6]) #get files with this processor, sort on starttime
      #detect if files are available multiple times (only difference is creation time... occurs in eg RPRO/CO/01_01/2018/06/17
      unique_files=set(map(lambda f: f.rsplit('_',1)[0],fs_p))
      if len(unique_files)!=len(fs_p): 
        for f in unique_files:
          #filter from fs_p files equal to f, sort on creation time and remove all files before that from the list
          toremove=sorted(filter(lambda _f: f==_f.rsplit('_',1)[0],fs_p),key=lambda ff: ff.split('_')[-1])[:-1]
          if toremove: logger.debug('Removing %d duplicates for %s'%(len(toremove),f))
          [fs_p.remove(_f) for _f in toremove]
      #determine the operations to execute: depends on target and processor version, and make sure all files are of the same type
      subjobs=[] #list of harpmerge commands, for non-sequential averaging of daily files (weekend days ... )
      logger.info('Checking %s,%s on %d files in %s'%(t,p,len(fs_p),_p))
      avg_level='TB' if len(processors_set)>1 else 'TD' #default avg_level for L2 input
      stime_mask='%Y%m%d*' #default mask on stime to recognize similar files for L2 input
      #get individual days from fs_p
      allstimes=list(map(lambda f: gettime(f.split('_')[-6]),fs_p));
      alldays=sorted(set(map(lambda _dt: _dt.date(),allstimes)))
      allmonths=sorted(set(map(getmonthbin,alldays)))
      allyears=sorted(set(map(getyearbin,allmonths)))
      logger.info('Found %d days, %d months and %d years'%(len(alldays),len(allmonths),len(allyears)))
      if not VITO and '_L2__' in t and len(alldays)!=1: logger.error('For the first step L2 -> L3 I expect orbits in day folders %s'%_p);continue
      #go through the possible cases, first the L2 inputs
      if all(map(lambda f: fnmatch.fnmatch(f,'S5P_*_L2__CO_*'),fs_p)): 
        tvar='CO_column_number_density'
        oper.extend(['derive(CO_column_volume_mixing_ratio_dry_air {time} [ppbv])',keep%('CO_column_volume_mixing_ratio_dry_air'),oper_bin])
        #add filter, depends on processor
        if p.split('_')[-1]>='010202':
          oper.insert(0,'CO_column_number_density_validity >=70') #mid level clouds
        else: 
          oper=['solar_zenith_angle <80','sensor_zenith_angle <65','cloud_height <5000','validity !& 4458240','scan_subindex >1']+oper #4458240 for SRON ... 1073739775 more strict, better at wollongong
        if int(p.split('_')[0])>1:
          if harpversion<(1,13): pass #raise ValueError('S5P CO L2 data contains destriped columns...update harp to read it')
          if 'S5P_NRTI_L2__' not in t: option='co=corrected'
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*_L2__NO2_*'),fs_p)):
        tvar='tropospheric_NO2_column_number_density'
        oper.extend([keep%(','.join((tvar,'scan_subindex','cloud_fraction'))),oper_bin])
        #oper=['solar_zenith_angle <80','tropospheric_NO2_column_number_density_validity>=50','cloud_fraction<=0.3','cloud_fraction>=0']+oper  -> replaced by qa_value
        oper=['tropospheric_NO2_column_number_density_validity>75']+oper
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*_L2__AER_LH_*'),fs_p)): 
        tvar='aerosol_height';
        oper.insert(0,'aerosol_height_validity >0')
        oper.extend([keep%(tvar+',aerosol_optical_depth'),oper_bin])
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*_L2__HCHO_*'),fs_p)):
        tvar='tropospheric_HCHO_column_number_density'
        oper.insert(0,'%s_validity >50'%tvar)
        oper.extend([keep%(tvar),oper_bin])
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*_L2__CH4_*'),fs_p)):
        tvar='CH4_column_volume_mixing_ratio_dry_air'
        option='ch4=bias_corrected'
        oper.extend([keep%(','.join((tvar,))),oper_bin])        #'surface_albedo','cloud_fraction' --> for Mahesh
        oper.insert(0,'CH4_column_volume_mixing_ratio_dry_air_validity >50')
      #treat the L3 files
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*L3TB*_*'),fs_p)):
        #oper.insert(0,'exclude(CH4_column_volume_mixing_ratio_dry_air_count,weight,latitude_bounds_weight,longitude_bounds_weight)')
        oper.insert(0,exclude_default)
        subjobs=[(oper,'TD','%Y%m%d*',(dayfilter,(day,))) for day in alldays]
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*L3TD*_*'),fs_p)):
        #create monthly files, weekend file, weekday file
        subjobs=[(oper,'TM','%Y%m*',(monthfilter,(month,))) for month in allmonths]
        #specialekes
        #oper.insert(0,'exclude(CH4_column_volume_mixing_ratio_dry_air_count,weight,latitude_bounds_weight,longitude_bounds_weight)')
        oper.insert(0,exclude_default)
        [subjobs.extend([(oper,'m5','%Y%m*',(weekdayfilter,(month,(5,)))),#weekend sat
                 (oper,'m6','%Y%m*',(weekdayfilter,(month,(6,)))),#weekend sun
                 (oper,'mw','%Y%m*',(weekdayfilter,(month,(0,1,2,3,4)))),#mid-week
                 (oper,'ms','%Y%m*',(weekdayfilter,(month,(5,6)))),#weekend 
                 ]) for month in allmonths] # -> this is in UT... maybe better to work with local time at high latitudes-> overkill? 
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*L3TM*_*'),fs_p)):
        #oper.insert(0,'exclude(CH4_column_volume_mixing_ratio_dry_air_count,weight,latitude_bounds_weight,longitude_bounds_weight)')
        oper.insert(0,exclude_default)
        subjobs=[(oper,'TY','%Y*',(yearfilter,(year,))) for year in allyears]
      elif all(map(lambda f: fnmatch.fnmatch(f,'S5P_*L3TY*_*') or fnmatch.fnmatch(f,'S5P_*L3m[56ws]*_*'),fs_p)):
        continue #max depth
      else: logger.error('Unknown/invalid import file list: %s'%', '.join(fs_p)[:50]);raise ValueError('I do not know how to process these files')
      if 'S5P_NRTI_L2__' in t: #fix overplapping scanlines 
        orbit_index=int(fs_p[0].split('_')[-4])
        if t[13:].strip('_') in ('CO','CH4'): oper.insert(0,'index>=2580' if orbit_index<9388 else 'index>=3340') # remove 12 overlapping scanlines from granules for NRTI, from orbit 9388 pixel size has changed...more overlapping scanlines... for CO, CH4 12*215=2580 or 16*215=3340
        elif t[13:].strip('_') in ('HCHO','NO2'): oper.insert(0,'index>=5400' if orbit_index<9388 else 'index>=7200') # remove 12 overlapping scanlines from granules for NRTI, from orbit 9388 pixel size has changed...more overlapping scanlines... 12*450 or 16*450
        else: raise ValueError('number of pixels in scanline depends on target: CO is 215 for NO2 450 ?')
      subjobs=subjobs or [(oper,avg_level,stime_mask,(_NOFILEFILTER,()))]
      for (oper,avg_level,stime_mask,(fs_p_filter,extraargs)) in subjobs:
        if not fnmatch.fnmatch(avg_level,avg_level_filter): logger.debug('Skipping job %s, does not match avg_level_filter %s'%(avg_level,avg_level_filter));continue
        #construct outputfilename
        fs_p_sub=list(filter(lambda f: fs_p_filter(*((f,)+extraargs)),fs_p)) #can contain mondays, weekendays according to definition of fs_p_filter, default is all files
        if len(fs_p_sub)==0: logger.debug('No matches for %s after filter'%avg_level);continue
        else:
          if len(fs_p_sub)<_AVERAGE_THRESHOLD.get(avg_level,0): logger.info('Skipping creation of %s file: not enough input files %d<%d'%(avg_level,len(fs_p_sub),_AVERAGE_THRESHOLD[avg_level]));continue
        filecountlist=map(lambda f: 1 if f[9:11]=='L2' else int(f.split('_')[-4],16),fs_p_sub)
        stimes,etimes,ifcount,sortedfs_p_sub=zip(*sorted(map(lambda fc: fc[0].split('_')[-6:-4]+[fc[1],fc[0]],zip(fs_p_sub,filecountlist)),key=lambda x:x[0])) #get stime,etime,individual file count,file from filenames
        filecount=sum(ifcount) #count the number of files in the overall bin
        if filecount>_MAXBINCOUNT: raise ValueError('The total number of files in this %s binning job (%d files) exceeds the maximum'%(avg_level,filecount))
        #check possible time interval overlap in input files for L3 
        stimes_dt=np.array(list(map(gettime,stimes)));etimes_dt=np.array(list(map(gettime,etimes)))
        for st,et,c,f in zip(stimes_dt,etimes_dt,ifcount,sortedfs_p_sub):
          if f[9:11]=='L2': continue #L2 files
          if ((stimes_dt==st)*(etimes_dt==et)*(ifcount==c)).sum()>1: raise ValueError('There are files in the average with equal time ranges and equal file counts: %s'%f)
        #determine output filename using idmatches
        newt,newp=idmatch(t,p)
        outf=fs_p_sub[0].split('_');outf[-6]=stimes[0];outf[-5]=etimes[-1];[outf.__setitem__(-2-i,_pp) for i,_pp in enumerate(newp.split('_')[::-1])];outf[-4]='{filecount}'
        outf[-1]=dt.datetime.now().strftime('%Y%m%dT%H%M%S.nc')
        fstime_dt=gettime(outf[-6]) #to format output_folder
        fetime_dt=gettime(outf[-5])
        outf='_'.join(outf)
        #orbit number becomes=#files in bin, four digits should be enough for yearly NRT binning 366(days)*12(orbits)*12(granules)
        outf=(newt+outf[_LENS5PID:]).format(filecount='%05x'%filecount,avg_level=avg_level)
        outfolder=fstime_dt.strftime(output_folder)
        fullout=os.path.join(outfolder,outf) #allow outputfolder with time formatting
        simoutf=outf.split('_');
        simoutf[-6]=fstime_dt.strftime(stime_mask);simoutf[-4]=simoutf[-5]='*';#remove stime(partly),etime,orbit
        simoutf[-1]='*.[n0]c' #remove creation time, 0c means that no pixels were left after harpmerge
        simoutf=os.path.join(outfolder,'_'.join(simoutf)) #simoutf will wildcard stime,etime,creation time; keep collection,proc version
        logger.debug('Matching already existing outputfile %s against %s'%(outf,simoutf))
        existingfiles=sorted(glob.glob(simoutf),key=lambda x: (int(x.rsplit('_',1)[0]==fullout.rsplit('_',1)[0]),x.rsplit('_',1)[-1])) #get similar files as outputfile, sort with equal filename,creation time... remove files with etime different from new outputfile
        
        #determine what to do: 
        #if there are existingfiles and the target output is already created previously (same filecount and time interval), skip the binning job if automode and remove the other similar files
        if existingfiles and existingfiles[-1].rsplit('_',1)[0]==fullout.rsplit('_',1)[0]: skip=automode;existingfiles=existingfiles[:-1] if automode else existingfiles
        else: skip=False
        #we now filter the existingfiles only if they are incident with the new time interval
        existingfiles=list(filter(lambda f: check_existing_time_window(f,fstime_dt,fetime_dt),existingfiles))
        if len(existingfiles): [os.remove(ef) for ef in existingfiles*int(not simulation)];logger.info('Removing %s'%(', '.join(existingfiles)))
        if skip: continue

        logger.debug('Writing file to %s'%outf)
        if not isinstance(oper,str): oper=';'.join(filter(lambda o: o,oper_box+oper)) #clean the list for empty strings, playing safe...
        logger.info('Working on %s,%s for subgroup %s'%(_p,t,p))
        fullfs=map(lambda f: os.path.join(filemap[f],f),fs_p_sub)
        postaction='ar' if harpversion>=(1,10) else 'ap' #from version 1.10, the binning is done more memory efficient
        harpcmd='harpmerge %s %s -%s "bin();exclude(latitude_bounds_weight,longitude_bounds_weight)"%s %s%s %s'%(('-o "%s"'%option if option else ''),('-a "%s"'%oper if oper else ''),postaction,(' -ap "%s"'%extra_post_actions if extra_post_actions else ''),extra_harp_args,' '.join(fullfs),fullout)
        if simulation: logger.info(harpcmd)
        else:
          if not os.path.isdir(os.path.dirname(fullout)): os.makedirs(os.path.dirname(fullout))
          if npool==0: 
            r=run_job(harpcmd) #sequential 
            err=_checkoutput(r,fullout)
          else: jobs.append(harpcmd)
  #parallel execution
  if not simulation and npool and jobs:
    npool=min(len(jobs),npool,_MAXCPU)
    logger.info('Processing %d jobs in a pool of %d child processes'%(len(jobs),npool))
    pool=multiprocessing.Pool(processes=npool)
    results=pool.imap(run_job,jobs)
    for i,harpcmd in enumerate(jobs): 
      err=0
      fullout=harpcmd.split()[-1]
      try: r=results.next()
      except Exception as e: logger.critical('Parallel execution of harpmerge for %s produced an exception: %s'%(harpcmd,repr(e)));
      except KeyboardInterrupt: logger.error('Stop requested by user');pool.terminate();break
      finally: err=_checkoutput(r,fullout)
    pool.close();pool.join()


#functions to interpret the command line arguments
def pos_int(n):
  n=int(n)
  if n<0:
    raise argparse.ArgumentTypeError('%d is not a positive integer'%n)
  return n;
def circle_divisor(x):
  x=float(x)
  X=180/x
  if int(X)<X:
    raise argparse.ArgumentTypeError('%f is not a divisor of 180'%x)
  if divmod(x*100,1)[-1]>0:
    raise argparse.ArgumentTypeError('too low value %f<0.01'%x)
  return x;
def eval_func(x):
  x=eval(x)
  if not isinstance(x,type(_NOFILEFILTER)):  raise argparse.ArgumentTypeError('%s is not a function'%x)
  return x;   
def eval_logger(x):
  x=eval(x)
  if not isinstance(x,type(logger)):  raise argparse.ArgumentTypeError('%s is not a logger'%x)
  return x;   
def eval_tuple(x):
  x=eval(x)
  if not isinstance(x,tuple):  raise argparse.ArgumentTypeError('%s is not a tuple'%x)
  return x;
      
def main():
  parser = argparse.ArgumentParser(
      formatter_class=argparse.RawTextHelpFormatter,
      description = """This is a wrapper around the HARP toolbox for binning of S5P files from L2 orbits/granules to L3 daily, monthly and yearly files""",
      epilog      = """Written by BIRA-IASB, contact bavo.langerock@aeronomie.be

example create TD files
Process L2 orbits with processor >=010202: OFFL and RPRO 
./terrascope3.py -d /bira-iasb/data/SATELLITE/S5P/TROPOMI/L2/ESA_HUB/OFL/CO/01_01 -o ~/projects/FTIR/validation/s5p/binned_files/S5P_test_t/%Y -n 5 -a -f "lambda f: f.split('_')[-2]>='010202'"
Fix processor changes in L3 db
./terrascope3.py -d ~/projects/FTIR/validation/s5p/binned_files/S5P_test_t -o ~/projects/FTIR/validation/s5p/binned_files/S5P_test_t/%Y -a -n 5 -f "lambda f: 'L3TB' in f"

Several runs to converge into a final state of the archive, averages are updated on the go...
./terrascope3.py -d ~/projects/FTIR/validation/s5p/binned_files/S5P_test_t -o ~/projects/FTIR/validation/s5p/binned_files/S5P_test_t/%Y -a

Binned files are: 
  TB (binned, intermediate if processor change in a single day);
  TD (daily)
  TM (monthly)
  TY (yearly)
  m5,m6=monthly binned Saterdays,Sundays respectively
  mw=monthly binned working days (days 0-4)
  ms=monthly binned weekend days (days 5-6)""")
  parser.add_argument('-d', '--data_folder', type=str, required=True, help='location of S5P L2/L3 files (sorted in daily folder Y/m/d)')
  parser.add_argument('-o', '--output_folder',    type=str, required=True, help='location of target files, will be formatted by datetime')
  parser.add_argument('-n', '--npool',      type=pos_int, required=False, help='# CPU\'s, parallize on days, 0=serial execution', default=0)
  parser.add_argument('-v', '--version',    action='version',         version='%(prog)s'+' %s'%__version__)
  parser.add_argument('-s', '--simulation', action='store_true', help='run in simulation mode')
  parser.add_argument('-a', '--automode', action='store_true', help='run in automode (detects already created files)')
  parser.add_argument('-V', '--VITO', action='store_true', help='run at vito (L2 data is stored in folders)')
  parser.add_argument('--latstep', type=circle_divisor, help='lat step',default=0.05)
  parser.add_argument('--lonstep', type=circle_divisor, help='lon step',default=0.05)
  parser.add_argument('-f', '--filefilter', type=eval_func, required=False, help='lambda function to filter the list of source files',default=_NOFILEFILTER)
  parser.add_argument('-l', '--logger', type=eval_logger, required=False, help='logger definition',default=operlogger,choices=[logger,testlogger,operlogger])
  parser.add_argument('--format', type=str, required=False, help='format option is passed onto harpmerge',default='netcdf',choices=['netcdf','hdf5'])
  parser.add_argument('--hdf5-compression', type=pos_int, required=False, help='hdf5 compression is passed onto harpmerge',default=0,choices=range(10))
  parser.add_argument('--box',    type=eval_tuple,default=(),required=False,help='use --box "(41.5,62.5,-94.5,-56.5)" to get L3 over a specific latlon box')
  parser.add_argument('--avg_level_filter',    type=str,default='*',required=False,help='use --avg_level_filter "TM" to only create monthly means out of the daily files')
  parser.add_argument('--extra_post_actions',    type=str,default='',required=False,help='use --extra_post_actions "derive(latitude {time,latitude});derive(longitude {time,longitude})", only for HARP>=1.10')
  args = parser.parse_args()
  run_folder(**vars(args))
  return 
  
if __name__ == '__main__':
  main()
  
