import argparse
import json
import os
import subprocess
import sys
from datetime import datetime, timedelta

from pyspark import SparkContext
from pyspark import SparkConf
print (os.environ['PYTHONPATH'])
sys.path.insert(0, os.path.abspath(os.path.join(os.path.dirname(__file__), '.')))
ppath = os.environ['PYTHONPATH']

os.environ['PYTHONPATH'] = ppath + ':' + os.path.abspath(os.path.join(os.path.dirname(__file__), '.'))
#os.environ['PYSPARK_PYTHON'] = '/usr/bin/python3.5'
print (os.environ['PYTHONPATH'])
from s5pbira.s5p_product import s5p_L3_product
from s5pbira.S5p_convert_nc_gtiff import convert_nc2tiff

import logging


pyspark_log = logging.getLogger('pyspark').setLevel(logging.ERROR)  # remove anoying pyspark messages
logger = logging.getLogger("py4j")
logger.setLevel(logging.INFO)
logger.addHandler(logging.StreamHandler())

DOCKER_VOLUMES={'/data/MTDA/CGS_S5P': {'bind': '/data/MTDA/CGS_S5P', 'mode': 'ro'},     #L2 inputs
                '/data/TERRASCOPE/s5p_dev': {'bind': '/data/TERRASCOPE/s5p', 'mode': 'rw'}}             #L3 inputs or outputs   #TODO DEV vs PROD

'''
######################## Examples ##################"
###CO TD OFFL ####
--input_dir /data/MTDA/CGS_S5P/CGS_S5P_L2_CO --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_NO2_TD_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_CO_TD_V1 --product_type CO --composit_type TD --processing_mode OFFL --start_date 20221101 --end_date 20221101 --version 200 --summary_file /tmp/my_sum_CO.json
### HCHO TD OFFL ####
--input_dir /data/MTDA/CGS_S5P/CGS_S5P_L2_HCHO --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_HCHO_TD_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_HCHO_TD_V1 --product_type HCHO --composit_type TD --processing_mode OFFL --start_date 20221101 --end_date 20221101 --version 200 --summary_file /tmp/my_sum_HCHO.json
### CH4 TD ##
--input_dir /data/MTDA/CGS_S5P/CGS_S5P_L2_CH4 --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TD_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_CH4_TD_V1 --product_type CH4 --composit_type TD --processing_mode OFFL --start_date 20221102 --end_date 20221102 --version 200 --summary_file /tmp/my_sum_CH4.json
## CH4 TM ###
--input_dir /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TD_V1 --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TM_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_CH4_TM_V1 --product_type CH4 --composit_type TM --processing_mode OFFL --start_date 20221102 --end_date 20221102 --version 200 --summary_file /tmp/my_sum_CH4.json
##CH4 TY##
--input_dir /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TD_V1 --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_CH4_TY_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_CH4_TY_V1 --product_type CH4 --composit_type TY --processing_mode OFFL --start_date 20221102 --end_date 20221102 --version 200 --summary_file /tmp/my_sum_CH4.json
###NO2 ##
--input_dir /data/MTDA/CGS_S5P/CGS_S5P_L2_NO2 --output_dir_nc /data/TERRASCOPE/s5p_dev/output_nc/L3_NO2_TD_V1 --output_dir_tif /data/TERRASCOPE/s5p_dev/output/L3_NO2_TD_V1 --product_type NO2 --composit_type TD --processing_mode OFFL --start_date 20201219 --end_date 20201219 --version 100 --summary_file /tmp/my_sum.json --cluster
'''

def _str2bool(v):
    import argparse
    try:
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        if v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
    except:
        raise argparse.ArgumentTypeError('Wrong value in _str2bool: '+str(v))

def parse_arguments():
    parser = argparse.ArgumentParser(prog='workflow_runner.py')

    ### INPUTS ###
    parser.add_argument('-i', '--input_dir', required=True,
                        help="input root dir without year/month: eg: /data/s5p/CGS_S5P/CGS_S5P_L2_NO2 or /data/TERRASCOPE/s5p/output_nc/L3_NO2_TD_V1 ")
    # parser.add_argument('-min', '--minimal_version', required=False,
    #                     default=None,
    #                     help='The minimal version of the input files to used in the output. Versions below this will be skipped eg:010300')

    #OR start_date and end_date OR date_lst(Prime) should be provided
    parser.add_argument('--start_date', required=False,
                        help='the first date of the period to be processed [yyyymmdd]',
                        default=None),
    parser.add_argument('--end_date', required=False,
                        help='the last date of the period to be processed [yyyymmdd]',
                        default=None),
    parser.add_argument('--date_lst_file', required=False, type=str,
                        help='a file containing all dates that needs to be processed [yyyymmdd]',
                        default=None),

    ## PROCESS PARAMETERS ###
    parser.add_argument("--product_type", metavar='CO, HCHO, ...',
                        required=True,
                        choices=['NO2', 'CH4', 'CO', 'HCHO'],
                        help='the chemical substance'),
    parser.add_argument('-t', "--composit_type", required=True,
                        choices=['TD', 'TM', 'TY'],
                        help='TD|TM|TY'),
    parser.add_argument('-m', "--processing_mode", required=True,
                        choices=['OFFL', 'NRTI'],
                        help='OFFL, NRTI'),
    parser.add_argument('-v', '--version', required=True,
                        help='Main product version 100|200|..., without prefix V')


    ### OUTPUTS ###
    parser.add_argument('--output_dir_nc', metavar='DIRECTORY', default=None,
                        required=True,
                        help='The root output_dir for creating nc-files(direct output of harpmerge)'),
    parser.add_argument('--output_dir_tif', metavar='DIRECTORY', default=None,
                        required=True,
                        help='The root output_dir for creating single band tif-files'),
    parser.add_argument('--summary_file', metavar='FILE', default=None,
                        required=False,
                        help='The filepath where the print the json summary to. Default off'),
    parser.add_argument('--overwrite', action='store_true',
                        help='overwrite files if they already exists.'),

    #PROCESSING FLAGS
    parser.add_argument('-s', '--simulation', action='store_true',
                        help='run in simulation mode, do not execute the harpmerge/gdal_translate(but print the command)')
    parser.add_argument('-c', '--cluster', action='store_true',
                        help='run on spark_cluster, if not local(default)'),
    parser.add_argument('-rpro', "--also_use_RPRO", type=_str2bool,
                        required=False,
                        default=False,
                        help='also use L2 RPRO-data in L3 OFFL. in case of double overpasses, use the RPRO-product only. ')

    #when running on a local machine we have to use docker by docker-py package. The following options are possible:
    # #In normal operations, this scripted will be called inside the docker container
    # #DOCKER OPTIONS
    # parser.add_argument('-d', '--docker_container', required=False,
    #                     default='vito-docker-private.artifactory.vgt.vito.be/s5p_harp1.16',
    #                     help='eg: vito-docker-private.artifactory.vgt.vito.be/s5p_harp1.16')
    # parser.add_argument('-dm', '--docker_hostvolume_mount', required=False,
    #                     default='/data/TERRASCOPE/s5p_dev',
    #                     help='eg: This host directory will be used as mount and will be accessible via '
    #                          '/data/TERRASCOPE/s5p. The value should be different between DEV and PROD')
    return parser.parse_args()

def process_single_product_hello_world(date_str):
    print("date_str: %s" % date_str)
    return date_str

def process_single_product(input_root_dir, output_dir_nc, output_dir_tif, process_date, product_type, composit_type, processing_mode, version, overwrite=True, simulation=False, use_local_docker=False, use_rpro=False):
    #use_local_docker : Set to True if you want this to run using docker exec (on a local ALREADY RUNNING docker container with name 's5p_harp'
    status = 0      #keep track of processing, it will be <0 on failure
    tif_files = []  # a list of resulting tif files

    logger = logging.getLogger()
    new_product = s5p_L3_product(input_root_dir, output_dir_nc, output_dir_tif, process_date, product_type, composit_type, processing_mode, version, use_rpro=use_rpro) #if minimal version is None, the version will be set based on the revision number(01->010000)
    nr_of_inputs = new_product.nr_of_inputs
    if new_product.status != 0:
        return new_product, new_product.status, nr_of_inputs, tif_files
    #new_product.create_output_folder(logger)
    product_id = new_product.title
    logger.info("Processing %s:" % product_id)
    cmd = 'harpmerge ' + new_product.harp_parameters
    cmd = cmd.replace("'", '"')
    cmd_docker = "docker exec s5p_harp bash -c '%s' " % cmd
    if use_local_docker:
        cmd = cmd_docker
    logger.info("Executing %s" %cmd)
    out_nc_exist = os.path.exists(os.path.join(new_product.outfolder_nc, new_product.outncfile))

    if not out_nc_exist or (out_nc_exist and overwrite):
        logger.info("Starting harpmerge command...")
        if simulation :
            logger.info("THIS IS A SIMULATION RUN")
            cmd = "echo %s" % cmd

        p = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
        stdout, stderr = p.communicate()
        errNr = p.returncode
        if errNr > 0:
            if str(stderr).find("merged product is empty") > 0:
                #this is not a failure, it is just a case of nodata
                logger.error(" ********* S5P harpmerge RESULT : NODATA(merged product is empty) on %s : %s" % (new_product.title, new_product.outncfile))
                status = 5  #NODATA(harpmerge)
            else:
                logger.error('***********************ERROR S5P RESULT ************************')
                logger.error("Product : %s : %s" % (new_product.title, new_product.outncfile))
                logger.error('errNr  : %s' % errNr)
                logger.error("stdout : %s" % stdout)
                logger.error("stderr : %s" % stderr)
                logger.error("harp command : %s" % cmd)
                logger.error("***********************************************************")

                status = -1 #FAILURE
                #raise Exception("Error while executing harpmerge %s %s %s " % (errNr, stdout, stderr))

        if status == 0:
            # create tiff files
            out_tiff_exist = os.path.exists(os.path.join(new_product.outfolder_tif, new_product.outncfile.replace(".nc", '.tif')))
            if not out_tiff_exist or (out_tiff_exist and overwrite):
                logger.info("Converting NC to GeoTiff: %s " % new_product.outncfile)
                nc_file = os.path.join(new_product.outfolder_nc, new_product.outncfile)
                if not simulation:
                    status, tif_files = convert_nc2tiff(nc_file, new_product.outfolder_tif, new_product.tif_settings)
                else:
                    logger.info("THIS IS A SIMULATION RUN")
                    status = 0
                    tif_files = ['simulation_tif_file', 'simulation_tif_File2']
                logger.info("DONE converting NC to GeoTiff: %s " % new_product.outncfile)
    else:
        logger.info("this is a simulation run, we didn't do anything")

    new_product.set_status(status)
    return new_product, status, nr_of_inputs, tif_files

def create_json_summary(new_product, tif_files):
    print("%s" % new_product)
    print("%s" % tif_files)
    return tif_files

def get_date_lst(start_date, end_date, composit_type='TD'):
    #create a list of dates between start_date and end_date(both included)
    date_lst = []
    yyyymmdd = start_date
    date_obj = datetime.strptime(yyyymmdd, '%Y%m%d')
    end_date_obj = datetime.strptime(end_date, '%Y%m%d')
    while date_obj <= end_date_obj:
        if composit_type == 'TD':
            date_str = date_obj.strftime('%Y%m%d')
            date_lst.append(date_str)
            date_obj += timedelta(days=1)
        elif composit_type == 'TM':
            date_str = '%s01' % date_obj.strftime('%Y%m')    #convert to first of the month
            date_lst.append(date_str)
            #adding 1 month, by adding 35 days and replace the day by 01
            date_obj += timedelta(days=35)
            date_str = '%s01' % date_obj.strftime('%Y%m')
            date_obj = datetime.strptime(date_str,"%Y%m%d")
            #date_obj += relativedelta(months=1)
        elif composit_type == 'TY':
            date_str = '%s0101' % date_obj.strftime('%Y')  # convert to first of the month
            date_lst.append(date_str)
            #date_obj += relativedelta(years=1)
            #adding one year by adding 370 days and set the day to 01
            date_obj += timedelta(days=370)
            date_str = '%s01' % date_obj.strftime('%Y%m')
            date_obj = datetime.strptime(date_str, "%Y%m%d")
    return date_lst
    #get_date_lst('20220101', '20220201', period=1)


def main():

    args = parse_arguments()

    #Process in dates arguments
    if args.date_lst_file is None:
        date_lst = get_date_lst(args.start_date, args.end_date, composit_type=args.composit_type)   #get all dates between start and end
    else:
        if os.path.exists(args.date_lst_file):
            with open(args.date_lst_file, 'r') as f:
                date_lst = f.readlines()
            date_lst = list(map(lambda yyyymmdd: yyyymmdd.strip(), date_lst))    #strip newlines
        else:
            raise Exception("date_lst_file does not exist on %s" % args.date_lst_file)
    nr_of_jobs = len(date_lst)

    if args.cluster:
        try:
            if nr_of_jobs > 0:
                sc_conf = SparkConf()
                sc = SparkContext(conf=sc_conf)
                sparkApplicationId = sc.applicationId
                logger.info("spark id : %s" % sparkApplicationId)
                logger.info("Simulation run : %s" % args.simulation)

                jobs_rdd = sc.parallelize(date_lst, nr_of_jobs)

                # print( '%s,%s,%s,%s,%s,%s,%s,%s,%s,%s' % args.input_dir, args.output_dir_nc, args.output_dir_tif,
                #                             date_lst[0], args.product_type, args.composit_type, args.processing_mode,
                #                             args.version, args.simulation, args.minimal_version)
                return_lst = jobs_rdd.map(lambda process_date:
                                             process_single_product(args.input_dir, args.output_dir_nc, args.output_dir_tif,
                                             process_date, args.product_type, args.composit_type, args.processing_mode,
                                             args.version, simulation=args.simulation, use_rpro=args.also_use_RPRO)).collect()
                # return_lst = jobs_rdd.map(lambda process_date: process_single_product_hello_world(process_date)).collect()
                print(return_lst)
                # jobs_rdd.foreach(lambda process_date: process_single_product(args.input_dir, args.output_dir_nc, args.output_dir_tif,
                #                             process_date, args.product_type, args.composit_type, args.processing_mode,
                #                             args.version, simulation=args.simulation, minimal_version=args.minimal_version,
                #                             logger=None))
        finally:
            sc.stop()
    else:
        #LOCAL PROCESSING
        #startup the docker container first!!!
        #docker run --name "s5p_harp" -v "/data/MTDA/CGS_S5P:/data/MTDA/CGS_S5P:ro" -v "/data/TERRASCOPE/s5p_dev:/data/TERRASCOPE/s5p_dev:rw" -idt vito-docker-private.artifactory.vgt.vito.be/s5p_harp1.16
        return_lst=[]
        for process_date in date_lst:
            #product_id_status will be a tuple of  (product_id, 0, tif_files)
            product_json_results = process_single_product(args.input_dir, args.output_dir_nc, args.output_dir_tif,
                                            process_date, args.product_type, args.composit_type, args.processing_mode,
                                            args.version, simulation=args.simulation,
                                            use_local_docker=True, use_rpro=args.also_use_RPRO)

            return_lst.append(product_json_results)


    ##### PRINT SUMMARY ####
    success_counter = 0
    total_counter = 0
    failure_counter = 0
    corrupt_pattern_counter = 0
    version_mismatch_counter = 0
    nodata_counter = 0  #for CH4 it can be possible that there is not valid data for a day, harpmerge generates the following message :WARNING: merged product is empty. We should handle this a slightly different way than FAILURE or OK
    product_lst = []
    for product, status, nr_of_inputs, tif_files in return_lst:
        product_dict = product.print_result()

        total_counter += 1
        if status < 0:
            failure_counter += 1
            product_dict['properties']['processing_status']['status'] = "FAILED"
        elif status == 3:
            version_mismatch_counter += 1
            product_dict['properties']['processing_status']['status'] = "VERSION MISMATCH"
        elif status == 4:
            corrupt_pattern_counter += 1
            product_dict['properties']['processing_status']['status'] = "No MATCHING PATTERN"
        elif status == 5:
            nodata_counter += 1
            product_dict['properties']['processing_status']['status'] = "NODATA"
        elif status == 0:
            success_counter +=1
        product_lst.append(product_dict)

    total_failure_counter = corrupt_pattern_counter + version_mismatch_counter + failure_counter
    print ('--------------------------------------')
    print ('- number of TOTAL jobs : %d' % total_counter)
    print ('- number of DONE jobs : %d (success)' % success_counter)
    print ('- number of FAILED jobs : %d (error during harpmerge)' % failure_counter)
    print ('- number of invalid PATTERN jobs : %d ( possible RPRO/OFFL/NRTI mismatch?)' % corrupt_pattern_counter)
    print ('- number of VERSION MISMATCH jobs : %d (different inputversions than expected 01 vs 02)' % version_mismatch_counter)
    print( '- number of NODATA jobs : %d (nodata detected by harpmerge)' % nodata_counter)

    summary_dict = {}       #output as json
    if args.summary_file:
        summary_dict.update(nr_of_jobs_total=total_counter)
        summary_dict.update(nr_of_succeeded=success_counter)
        summary_dict.update(nr_of_failures=failure_counter)
        summary_dict.update(nr_of_pattern_mismatch=corrupt_pattern_counter)
        summary_dict.update(nr_of_version_mismatch=version_mismatch_counter)
        summary_dict.update(jobs=product_lst)

        print('- summary file : %s' % args.summary_file)
        out_file = open(args.summary_file, 'w')
        json.dump(summary_dict, out_file, indent=4)
        out_file.close()

    if total_failure_counter > 0:
        return -1

    return 0

    # #loop over all possible jobs(days)
    # try:
    #
    #     # prepare jobs, set the metadata
    #     out_joblst = []
    #     for job in in_joblst:
    #         s5p_prod = s5p_out_product(job, logger, args.version, args.composit_type)
    #         s5p_prod.update_metadata_json()
    #         out_joblst.append(s5p_prod)
    #
    #
    #     #execute jobs in parallel
    #     numberOfJobs = len(out_joblst)
    #     if numberOfJobs > 0:
    #         jobs_rdd = sc.parallelize(out_joblst, len(out_joblst))
    #         jobs_rdd.foreach(lambda job: process(job, logger, args.simulation))
    #
    #     # calcuate simple footprint and put this in the job.content dictionary
    #     json_content = list(map(lambda job: job.set_metadata("footprint", get_footprint_value(
    #         os.path.join(job.content['outdir_tif'],
    #                      job.content['output_tif_list'][0]))), out_joblst))
    #
    #     #write all output to a json file
    #     json_content = list(map(lambda job : job.content, out_joblst))
    #     with open(out_json, 'w') as f:
    #         json.dump(json_content, f, indent=4)
    #     logger.info("writing output json %s" % out_json)



if __name__ == '__main__':
    status = main()
    sys.exit(status)